# A simple echo server 

from curio import run, tcp_server
from curio.socket import IPPROTO_TCP, TCP_NODELAY

async def echo_handler(client, addr):
    print('Connection from', addr)
    client.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
    while True:
        data = await client.recv(1000000)
        if not data:
            break
        await client.sendall(data)
    print('Connection closed')

if __name__ == '__main__':
    run(tcp_server('', 25000, echo_handler))



# client.py
#
# Common client. Measure the response rate of the echo server
# Sends a large number of requests. Measures how long it takes.

from socket import *
import time
import sys

if len(sys.argv) > 1:
    MSGSIZE = int(sys.argv[1])
else:
    MSGSIZE = 1000

msg = b'x'*MSGSIZE

def run_test(n):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.connect(('localhost', 25000))
    while n > 0:
        sock.sendall(msg)
        nrecv = 0
        while nrecv < MSGSIZE:
            resp = sock.recv(MSGSIZE)
            if not resp:
                raise SystemExit()
            nrecv += len(resp)
        n -= 1

NMESSAGES = 1000000
print('Sending', NMESSAGES, 'messages')
start = time.time()
run_test(NMESSAGES)
end = time.time()
duration = end-start
print(NMESSAGES,'in', duration)
print(NMESSAGES/duration, 'requests/sec')



# echoserv.py
#
# Echo server using streams

from curio import run, tcp_server

async def echo_client(client, addr):
    print('Connection from', addr)
    s = client.as_stream()
    async for line in s:
        await s.write(line)
    print('Connection closed')
    await s.close()

if __name__ == '__main__':
    run(tcp_server('', 25000, echo_client))



# test_kernel.py

import time
from curio import *

def test_hello(kernel):
    results = []
    async def hello():
        results.append('hello')

    kernel.run(hello())
    assert results == [ 'hello' ]

def test_sleep(kernel):
    results = []
    async def main():
          results.append('start')
          await sleep(0.5)
          results.append('end')

    start = time.time()
    kernel.run(main())
    end = time.time()
    assert results == [
            'start',
            'end',
            ]
    elapsed = end-start
    assert elapsed > 0.5

def test_sleep_cancel(kernel):
    results = []

    async def sleeper():
        results.append('start')
        try:
            await sleep(1)
            results.append('not here')
        except CancelledError:
            results.append('cancelled')

    async def main():
        task = await spawn(sleeper())
        await sleep(0.5)
        await task.cancel()

    kernel.run(main())
    assert results == [
            'start',
            'cancelled',
            ]


def test_sleep_timeout(kernel):
    results = []

    async def sleeper():
        results.append('start')
        try:
            await timeout_after(0.5, sleep(1))
            results.append('not here')
        except TaskTimeout:
            results.append('timeout')

    async def main():
        task = await spawn(sleeper())
        await task.join()

    kernel.run(main())
    assert results == [
            'start',
            'timeout',
            ]

def test_sleep_ignore_timeout(kernel):
    results = []

    async def sleeper():
        results.append('start')
        if await ignore_after(0.5, sleep(1)) is None:
            results.append('timeout')

        async with ignore_after(0.5) as s:
            await sleep(1)

        if s.result is None:
            results.append('timeout2')


    async def main():
        task = await spawn(sleeper())
        await task.join()

    kernel.run(main())
    assert results == [
            'start',
            'timeout',
            'timeout2',
            ]

def test_sleep_notimeout(kernel):
    results = []

    async def sleeper():
        results.append('start')
        try:
            await timeout_after(1.5, sleep(1))
            results.append('here')
        except TaskTimeout:
            results.append('not here')

        await sleep(1)
        results.append('here2')

    async def main():
        task = await spawn(sleeper())
        await task.join()

    kernel.run(main())
    assert results == [
            'start',
            'here',
            'here2'
            ]

def test_task_join(kernel):
    results = []

    async def child():
        results.append('start')
        await sleep(0.5)
        results.append('end')
        return 37

    async def main():
        task = await spawn(child())
        await sleep(0.1)
        results.append('joining')
        r = await task.join()
        results.append(r)

    kernel.run(main())
    assert results == [
            'start',
            'joining',
            'end',
            37
            ]

def test_task_join_error(kernel):
    results = []

    async def child():
        results.append('start')
        int('bad')

    async def main():
        task = await spawn(child())
        await sleep(0.1)
        results.append('joining')
        try:
            r = await task.join()
            results.append(r)
        except TaskError as e:
            results.append('task fail')
            results.append(type(e))
            results.append(type(e.__cause__))

    kernel.run(main())
    assert results == [
            'start',
            'joining',
            'task fail',
            TaskError,
            ValueError,
            ]

def test_task_cancel(kernel):
    results = []

    async def child():
        results.append('start')
        try:
            await sleep(0.5)
            results.append('end')
        except CancelledError:
            results.append('cancelled')

    async def main():
        task = await spawn(child())
        results.append('cancel start')
        await sleep(0.1)
        results.append('cancelling')
        await task.cancel()
        results.append('done')

    kernel.run(main())
    assert results == [
            'start',
            'cancel start',
            'cancelling',
            'cancelled',
            'done',
            ]


def test_task_cancel_join(kernel):
    results = []

    async def child():
        results.append('start')
        await sleep(0.5)
        results.append('end')

    async def main():
        task = await spawn(child())
        results.append('cancel start')
        await sleep(0.1)
        results.append('cancelling')
        await task.cancel()
        # Try joining with a cancelled task. Should raise a TaskError
        try:
            await task.join()
        except TaskError as e:
            if type(e.__cause__) == CancelledError:
                results.append('join cancel')
            else:
                results.append(str(e.__cause__))
        results.append('done')

    kernel.run(main())
    assert results == [
            'start',
            'cancel start',
            'cancelling',
            'join cancel',
            'done',
            ]


def test_task_cancel_join_wait(kernel):
    results = []

    async def child():
        results.append('start')
        await sleep(0.5)
        results.append('end')

    async def canceller(task):
        await sleep(0.1)
        results.append('cancel')
        await task.cancel()

    async def main():
        task = await spawn(child())
        results.append('cancel start')
        await spawn(canceller(task))
        try:
            results.append('join')
            await task.join()     # Should raise TaskError... with CancelledError as cause
        except TaskError as e:
            if type(e.__cause__) == CancelledError:
                results.append('join cancel')
            else:
                results.append(str(e.__cause__))
        results.append('done')

    kernel.run(main())
    assert results == [
            'start',
            'cancel start',
            'join',
            'cancel',
            'join cancel',
            'done',
            ]

def test_task_child_cancel(kernel):
    results = []

    async def child():
        results.append('start')
        try:
            await sleep(0.5)
            results.append('end')
        except CancelledError:
            results.append('child cancelled')

    async def parent():
        try:
             child_task = await spawn(child())
             await sleep(0.5)
             results.append('end parent')
        except CancelledError:
            await child_task.cancel()
            results.append('parent cancelled')

    async def grandparent():
        try:
            parent_task = await spawn(parent())
            await sleep(0.5)
            results.append('end grandparent')
        except CancelledError:
            await parent_task.cancel()
            results.append('grandparent cancelled')

    async def main():
        task = await spawn(grandparent())
        await sleep(0.1)
        results.append('cancel start')
        await sleep(0.1)
        results.append('cancelling')
        await task.cancel()
        results.append('done')

    kernel.run(main())

    assert results == [
            'start',
            'cancel start',
            'cancelling',
            'child cancelled',
            'parent cancelled',
            'grandparent cancelled',
            'done',
            ]

def test_task_ready_cancel(kernel):
    # This tests a tricky corner case of a task cancelling another task that's also 
    # on the ready queue.
    results = []

    async def child():
        try:
            results.append('child sleep')
            await sleep(1.0)
            results.append('child slept')
            await sleep(1.0)
            results.append('should not see this')
        except CancelledError:
            results.append('child cancelled')

    async def parent():
        task = await spawn(child())
        results.append('parent sleep')
        await sleep(0.5)
        results.append('cancel start')
        await task.cancel()
        results.append('cancel done')

    async def main():
        task = await spawn(parent())
        await sleep(0.1)
        time.sleep(1)      # Forced block of the event loop. Both tasks should awake when we come back
        await sleep(0.1)

    kernel.run(main())

    assert results == [
            'child sleep',
            'parent sleep',
            'cancel start',
            'child slept',
            'child cancelled',
            'cancel done'
            ]


def test_double_cancel(kernel):
    results = []

    async def sleeper():
        results.append('start')
        try:
            await sleep(1)
            results.append('not here')
        except CancelledError:
            results.append('cancel request')
            await sleep(1)
            results.append('cancelled')

    async def main():
        task = await spawn(sleeper())
        await sleep(0.5)
        try:
            await timeout_after(1, task.cancel())
        except TaskTimeout:
            results.append('retry')
            await task.cancel()    # This second cancel should not abort any operation in sleeper
            results.append('done cancel')

    kernel.run(main())
    assert results == [
            'start',
            'cancel request',
            'retry',
            'cancelled',
            'done cancel'
            ]




# udp_echo
#
# An example of a simple UDP echo server. 

import curio
from curio import socket

async def main(addr):
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.bind(addr)
    while True:
        data, addr = await sock.recvfrom(10000)
        print('Received from', addr, data)
        await sock.sendto(data, addr)

if __name__ == '__main__':
    curio.run(main(('', 26000)))



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(name = "curio",
            description="Curio - Concurrent I/O",
            long_description = """
Curio is a library for performing concurrent I/O with coroutines in Python 3. 
""",
            license="""BSD""",
            version = "0.2",
            author = "David Beazley",
            author_email = "dave@dabeaz.com",
            maintainer = "David Beazley",
            maintainer_email = "dave@dabeaz.com",
            url = "https://github.com/dabeaz/curio",
            packages = ['curio'],
            classifiers = [
              'Programming Language :: Python :: 3',
              ]
            )



# Example: A simple echo server written using streams

from curio import run, spawn
from curio.socket import *

async def echo_server(address):
    sock = socket(AF_INET, SOCK_STREAM)
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
    sock.bind(address)
    sock.listen(5)
    print('Server listening at', address)
    async with sock:
        while True:
             client, addr = await sock.accept()
             print('Connection from', addr)
             await spawn(echo_client(client))

async def echo_client(client):
    s = client.as_stream()
    async for line in s:
        await s.write(line)
    await client.close()
    print('Connection closed')

if __name__ == '__main__':
     run(echo_server(('',25000)))



# curio/queue.py
#
# Implementation of a queue object that can be used to communicate
# between tasks.  This is only safe to use within curio. It is not
# thread-safe.

__all__ = [ 'Queue' ]

from collections import deque

from .traps import _wait_on_queue, _reschedule_tasks
from .kernel import kqueue

class Queue(object):
    __slots__ = ('maxsize', '_queue', '_get_waiting', 
                 '_put_waiting', '_join_waiting', '_task_count')

    def __init__(self, maxsize=0):
        self.maxsize = maxsize
        self._queue = deque()
        self._get_waiting = kqueue()
        self._put_waiting = kqueue()
        self._join_waiting = kqueue()
        self._task_count = 0

    def empty(self):
        return not self._queue

    def full(self):
        return self.maxsize and len(self._queue) == self.maxsize

    async def get(self):
        if self.empty():
            await _wait_on_queue(self._get_waiting, 'QUEUE_GET')
        result = self._queue.popleft()
        if self._put_waiting:
            await _reschedule_tasks(self._put_waiting, n=1)
        return result

    async def join(self):
        if self._task_count > 0:
            await _wait_on_queue(self._join_waiting, 'QUEUE_JOIN')

    async def put(self, item):
        if self.full():
            await _wait_on_queue(self._put_waiting, 'QUEUE_PUT')
        self._queue.append(item)
        self._task_count += 1
        if self._get_waiting:
            await _reschedule_tasks(self._get_waiting, n=1)

    def qsize(self):
        return len(self._queue)

    async def task_done(self):
        self._task_count -= 1
        if self._task_count == 0 and self._join_waiting:
            await _reschedule_tasks(self._join_waiting, n=len(self._join_waiting))



from curio import run, spawn, tcp_server
from socket import *

async def echo_handler(client, addr):
    print('Connection from', addr)
    try:
        client.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)
    except (OSError, NameError):
        pass
    s = client.as_stream()
    while True:
        data = await s.read(102400)
        if not data:
            break
        await s.write(data)
    await s.close()
    print('Connection closed')

if __name__ == '__main__':
    run(tcp_server('', 25000, echo_handler))



# curio/__init__.py

__version__ = '0.2'

from .errors import *
from .task import *
from .signal import *
from .kernel import *
from .sync import *
from .queue import *
from .workers import *
from .network import *

__all__ = [ *errors.__all__,
            *task.__all__,
            *signal.__all__,
            *kernel.__all__,
            *sync.__all__,
            *queue.__all__,
            *workers.__all__,
            *network.__all__
             ]




# Example of launching a subprocess and reading streaming output

from curio import subprocess
import curio

async def main():
    p = subprocess.Popen(['ping', 'www.python.org'], stdout=subprocess.PIPE)
    async for line in p.stdout:
        print('Got:', line.decode('ascii'), end='')

if __name__ == '__main__':
    curio.run(main())



# client.py
#
# Common client. Measure the response rate of the echo server

from socket import *
import time
from threading import Thread
import atexit
import sys

if len(sys.argv) > 1:
    MSGSIZE = int(sys.argv[1])
else:
    MSGSIZE = 1

msg = b'x'*MSGSIZE

sock = socket(AF_INET, SOCK_STREAM)
sock.connect(('localhost', 25000))
sock.setsockopt(IPPROTO_TCP, TCP_NODELAY, 1)

N = 0
results = []
def monitor():
    global N
    while True:
        time.sleep(1)
        print(N, 'requests/sec')
        results.append(N)
        N = 0

Thread(target=monitor, daemon=True).start()

def print_average():
    import statistics
    print('Average', statistics.mean(results), 'requests/sec')
atexit.register(print_average)

while True:
    sock.sendall(msg)
    nrecv = 0
    while nrecv < MSGSIZE:
        resp = sock.recv(MSGSIZE)
        if not resp:
            raise SystemExit()
        nrecv += len(resp)
    N += 1



# curio/errors.py
# 
# Curio specific exceptions

__all__ = [
    'CurioError', 'CancelledError', 'TaskTimeout', 'TaskError'
    ]

class CurioError(Exception):
    pass

class CancelledError(CurioError):
    pass

class TaskTimeout(CurioError):
    pass

class TaskError(CurioError):
    pass

class _CancelRetry(Exception):
    pass




# test_network.py

from curio import *
from curio.socket import *

def test_tcp_echo(kernel):
    results = []

    async def handler(client, addr):
        results.append('handler start')
        while True:
            results.append('recv wait')
            data = await client.recv(100)
            if not data:
                break
            results.append(('handler', data))
            await client.sendall(data)
        results.append('handler done')

    async def client(address, serv):
        results.append('client start')
        sock = socket(AF_INET, SOCK_STREAM)
        await sock.connect(address)
        await sock.send(b'Msg1')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        await sock.send(b'Msg2')
        await sleep(0.1)
        resp = await sock.recv(100)
        results.append(('client', resp))
        results.append('client close')
        await sock.close()
        await serv.cancel()

    async def main():
       serv = await spawn(tcp_server('',25000,handler))
       await spawn(client(('localhost', 25000), serv))

    kernel.run(main())

    assert results == [
            'client start',
            'handler start',
            'recv wait',
            ('handler', b'Msg1'),
            'recv wait',
            ('client', b'Msg1'),
            ('handler', b'Msg2'),
            'recv wait',
            ('client', b'Msg2'),
            'client close',
            'handler done'
            ]



# curio/workers.py
#
# Functions for performing work outside of curio.  This includes
# running functions in threads, processes, and executors from the
# concurrent.futures module.

__all__ = [ 'run_in_executor', 'run_in_thread', 'run_in_process' ]

import multiprocessing
import threading

from .errors import CancelledError, TaskTimeout
from .traps import _future_wait, _get_kernel, _get_current
from . import sync
from .channel import Channel

async def run_in_executor(exc, callable, *args, **kwargs):
    '''
    Run callable(*args, **kwargs) in an executor such as
    ThreadPoolExecutor or ProcessPoolExecutor from the
    concurrent.futures module.  Be aware that on cancellation, any
    worker thread or process that was handling the request will
    continue to run to completion as a kind of zombie-- possibly
    rendering the executor unusable for subsequent work.

    This function is provided for compatibility with
    concurrent.futures, but is not the recommend approach for running
    blocking or cpu-bound work in curio. Use the run_in_thread() or
    run_in_process() methods instead.
    '''
    future = exc.submit(callable, *args, **kwargs)
    await _future_wait(future)
    return future.result()

MAX_WORKER_THREADS = 64

async def run_in_thread(callable, *args, **kwargs):
    '''
    Run callable(*args) in a separate thread and return the result.
    It's not possible to terminate threads so if a cancellation
    occurs, the associated worker thread will continue to run until
    completion as a kind of zombie.  When this happens, the cancelled
    worker will be immediately replaced by a new worker thread that's
    able to process further requests.  The cancelled worker thread
    will terminate when the callable completes its work (the result is
    discarded in that case).

    Note: It is advisable that all callables submitted to threads
    have a bound on their execution time (e.g., timeout or some 
    other mechanism).
    '''
    kernel = await _get_kernel()
    if not kernel._thread_pool:
        kernel._thread_pool = WorkerPool(ThreadWorker, MAX_WORKER_THREADS)
    return await kernel._thread_pool.apply(callable, args, kwargs)

MAX_WORKER_PROCESSES = multiprocessing.cpu_count()

async def run_in_process(callable, *args, **kwargs):
    '''
    Run callable(*args, **kwargs) in a separate process and return the
    result.  In the event of cancellation, the worker process is
    immediately terminated.

    The worker process is created using multiprocessing.Process().
    Communication with the process uses multiprocessing.Pipe() and an
    asynchronous message passing channel.  All function arguments and
    return values are seralized using the pickle module.  When
    cancelled, the Process.terminate() method is used to kill the
    worker process.  This results in a SIGTERM signal being sent to
    the process.

    The worker process is a separate isolated Python interpreter.
    Nothing should be assumed about its global state including shared
    variables, files, or connections.
    '''
    kernel = await _get_kernel()
    if not kernel._process_pool:
        kernel._process_pool = WorkerPool(ProcessWorker, 
                                          MAX_WORKER_PROCESSES)

    return await kernel._process_pool.apply(callable, args, kwargs)

# The _FutureLess class is a custom "Future" implementation solely for
# use by curio. It is used by the ThreadWorker class below and
# provides only the minimal set of functionality needed to transmit a
# result back to the curio kernel.  Unlike the normal Future class,
# this version doesn't require any thread synchronization or
# notification support.  By eliminating that, the overhead associated
# with the handoff between curio tasks and threads is substantially
# faster.
class _FutureLess(object):
    __slots__ = ('_callback', '_exception', '_result')

    def set_result(self, result):
        self._result = result
        self._callback(self)

    def set_exception(self, exc):
        self._exception = exc
        self._callback(self)

    def result(self):
        try:
            return self._result
        except AttributeError:
            return self._exception
        
    def add_done_callback(self, func):
        self._callback = func

    def cancel(self):
        pass

# A ThreadWorker represents a thread that performs work on behalf of a
# curio task.   A curio task initiates work by executing the
# apply() method. This passes the request to a background thread that
# executes it.  While this takes place, the curio task blocks, waiting
# for a result to be set on an internal Future.

class ThreadWorker(object):
    '''
    Worker that executes a callable on behalf of a curio task in a separate thread.
    '''
    def __init__(self):
        self.thread = None
        self.start_evt = None
        self.lock = None
        self.request = None

    def _launch(self):
        self.start_evt = threading.Event()
        self.thread = threading.Thread(target=self.run_worker, daemon=True)
        self.thread.start()

    def run_worker(self):
        while True:
            self.start_evt.wait()
            self.start_evt.clear()
            # If there is no pending request, but we were signalled to
            # start, it means terminate.
            if not self.request:
                return

            func, args, kwargs, future = self.request

            try:
                result = func(*args, **kwargs)
                future.set_result(result)
            except Exception as e:
                future.set_exception(e)

    def shutdown(self):
        self.request = None
        if self.start_evt:
            self.start_evt.set()

    async def apply(self, func, args=(), kwargs={}):
        '''
        Run the callable func in a separate thread and return the result.
        '''
        if self.thread is None:
            self._launch()

        # Set up a request for the worker thread
        future = _FutureLess()
        self.request = (func, args, kwargs, future)

        # Wait for the result to be computed.  Important: The
        # start_evt passed as an argument is used to make the
        # worker thread start processing.
        await _future_wait(future, self.start_evt)
        return future.result()

class ProcessWorker(object):
    '''
    Managed process worker for running CPU-intensive tasks.  The main
    purpose of this class is to run workers with reliable
    cancellation/timeout semantics. Specifically, if a worker is
    cancelled, the underlying process is also killed.   This, as
    opposed to having it linger on running until work is complete.
    '''

    def __init__(self):
        self.process = None
        self.client_ch = None

    def _launch(self):
        client_ch, server_ch = multiprocessing.Pipe()
        self.process = multiprocessing.Process(target=self.run_server, args=(server_ch,), daemon=True)
        self.process.start()
        server_ch.close()
        self.client_ch = Channel.from_Connection(client_ch)

    def shutdown(self):
        if self.process:
            self.process.terminate()
            self.process = None
            self.nrequests = 0

    def run_server(self, ch):
        while True:
            func, args, kwargs = ch.recv()
            try:
                result = func(*args, **kwargs)
                ch.send((True, result))
            except Exception as e:
                ch.send((False, e))
            func = args = kwargs = None
    
    async def apply(self, func, args=(), kwargs={}):
        if self.process is None or not self.process.is_alive():
            self._launch()

        msg = (func, args, kwargs)
        await self.client_ch.send(msg)
        success, result = await self.client_ch.recv()
        if success:
            return result
        else:
            raise result

# Pool of workers for carrying out jobs on behalf of curio tasks.
#
# A tricky situation arises when task cancellations/timeouts occur.
# The apply() method used by curio tasks is able to abort and return
# control without computing a result.  However, the background worker
# might continue to run until its task is complete.  When this happens
# the shutdown() method is triggered on the worker and it is removed
# from further service--to be replaced by a fresh worker.  For process
# workers, this results in a process termination.  For threads, the
# thread will continue to run until it completes, at which point it
# will terminate.

class WorkerPool(object):
    def __init__(self, workercls, nworkers):
        self.nworkers = sync.Semaphore(nworkers)
        self.workercls = workercls
        self.workers = [workercls() for n in range(nworkers)]
        
    def shutdown(self):
        for worker in self.workers:
            worker.shutdown()
        self.workers = None

    async def apply(self, func, args=(), kwargs={}):
         async with self.nworkers:
             worker = self.workers.pop()
             try:
                 return await worker.apply(func, args, kwargs)
             except CancelledError as e:
                 # If a worker is timed out or cancelled, we shut it
                 # down and replace it with a fresh worker.
                 worker.shutdown()
                 worker = self.workercls()
                 raise
             finally:
                 self.workers.append(worker)

# Pool definitions should anyone want to use them directly
ProcessPool = lambda nworkers: WorkerPool(ProcessWorker, nworkers)
ThreadPool = lambda nworkers: WorkerPool(ThreadWorker, nworkers)



# An example of a server involving a CPU-intensive task.  We'll farm the 
# CPU-intensive work out to a separate process.

from curio import run, run_in_process, tcp_server

def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n-1) + fib(n-2)

async def fib_handler(client, addr):
    print('Connection from', addr)
    s = client.as_stream()
    async for line in s:
        try:
            n = int(line)
            result = await run_in_process(fib, n)
            resp = str(result) + '\n'
            await s.write(resp.encode('ascii'))
        except ValueError:
            await s.write(b'Bad input\n')
    print('Connection closed')
    await client.close()

if __name__ == '__main__':
    run(tcp_server('', 25000, fib_handler))



    
